/*
  gpsinfo.cpp
  Copyright (C) 2008 Tomohiro Masubuchi <bucchi@e-ml.jp>

  This file is part of qgmap.

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; If not, see <http://www.gnu.org/licenses/>.
*/

#include "qgmap.h"
#include "gpsinfo.h"
#include "pos.h"

#include <qstring.h>
#include <qstringlist.h>
#include <qdatetime.h>

#include <pthread.h>
#include <string.h>

GPSInfo::GPSInfo(void)
{
	pthread_mutex_init(&lock, NULL);
	is_updated = false;
}

GPSInfo GPSInfo::copy()
{
	pthread_mutex_lock(&lock);

	GPSInfo tmp;
	memcpy(&tmp, this, sizeof(*this));
	is_updated = false;

	pthread_mutex_unlock(&lock);

	memset(&tmp.lock, 0, sizeof(tmp.lock));
	pthread_mutex_init(&tmp.lock, NULL);

	return tmp;
}

GPSInfo::~GPSInfo()
{
	pthread_mutex_destroy(&lock);
}

// getter

#define SafetyCopy(item) ({						\
		typeof(item) tmp;						\
		pthread_mutex_lock(&lock);				\
		tmp = item;								\
		pthread_mutex_unlock(&lock);			\
		tmp; })


QDateTime GPSInfo::getDateTime() { return SafetyCopy(datetime); }
double GPSInfo::getLatitude() { return SafetyCopy(latitude); }
double GPSInfo::getLongitude() { return SafetyCopy(longitude); }
double GPSInfo::getAltitude() { return SafetyCopy(altitude); }
int GPSInfo::getSatellites() { return SafetyCopy(satellites); }
double GPSInfo::getVelocity() { return SafetyCopy(velocity); }
double GPSInfo::getDegree() { return SafetyCopy(degree); }
bool GPSInfo::isUpdated() { return SafetyCopy(is_updated); }


void GPSInfo::updateInfo(QString const &nmea)
{
	QString type;
	QStringList args;

	if(nmea.mid(0,3) != "$GP")
		return;

	type = nmea.mid(3,3);
	args = QStringList::split(",", nmea.mid(7));

	if(type == "GGA" && args.count() > 5){
		int state = args[5].toInt();

		pthread_mutex_lock(&lock);

		if (state > 0) {
			datetime.setTime(decodeNMEATime(args[0]));
			latitude = decodeNMEALatitude(args[1], args[2]);
			longitude = decodeNMEALongitude(args[3], args[4]);
		}
		satellites = args[6].toInt();
		status = state;
		is_updated = true;

		pthread_mutex_unlock(&lock);

	} else if(type == "RMC" && args.count() > 9){

		if(args[1] == "A"){
			int y, m, d, tmp;
			QTime time;
			QDate date;

			pthread_mutex_lock(&lock);

			velocity = args[6].toDouble() * 1.852;
			degree = args[7].toDouble();

			time = decodeNMEATime(args[0]);

			tmp = args[8].toInt();
			y = tmp % 100;
			m = (tmp / 100) % 100;
			d = (tmp / 10000) % 100;

			date.setYMD(2000 + y, m, d);

			datetime.setDate(date);
			datetime.setTime(time);

			is_updated = true;
			pthread_mutex_unlock(&lock);
		}

	}
}

QTime GPSInfo::decodeNMEATime(QString const &time)
{
	int h, m, s, ms = 0;

	h = time.mid(0, 2).toInt();
	m = time.mid(2, 2).toInt();
	s = time.mid(4, 2).toInt();
	if(time.length() == 9 && time[6] == '.')
		ms = time.mid(7, 2).toInt() * 10;

	return QTime(h, m, s, ms);
}

double GPSInfo::decodeNMEALatitude(QString const &lat_str,
									 QString const &dir)
{
	int min_pos = lat_str.find('.') - 2;
	double lat;

	if(min_pos < 0)
		return 0.0;

	lat = lat_str.mid(min_pos).toDouble() / 60.0;
	lat += lat_str.mid(0, min_pos).toDouble();

	if(dir == "S"){
		lat = -lat;
	} else if(dir != "N"){
		return 0.0;
	}

	return lat;
}

double GPSInfo::decodeNMEALongitude(QString const &lon_str,
									  QString const &dir)
{
	int min_pos = lon_str.find('.') - 2;
	double lon;

	if(min_pos < 0)
		return 0.0;

	lon = lon_str.mid(min_pos).toDouble() / 60.0;
	lon += lon_str.mid(0, min_pos).toDouble();

	if(dir == "W"){
		lon = -lon;
	} else if(dir != "E"){
		return 0.0;
	}

	return lon;
}

void GPSInfo::convTokyoToWGS()
{
	double lat;
	double lon;

	pthread_mutex_lock(&lock);

	Pos::convTokyoToWGS(&lat, &lon, latitude, longitude);
	latitude = lat;
	longitude = lon;

	pthread_mutex_unlock(&lock);
}

void GPSInfo::setTimeZone(int timezone)
{
	pthread_mutex_lock(&lock);

	datetime = datetime.addSecs(timezone * 60 * 60);

	pthread_mutex_unlock(&lock);
}
